/***************************************************************************
 *  BpmDj: Free Dj Tools
 * -------------------------------------------------------------------------
 *  Copyright (C) 2004 by Werner Van Belle, see copyright.txt
 *  e-mail: werner.van.belle@itek.norut.no                                
 ***************************************************************************/

#ifndef CLUSTER_H
#define CLUSTER_H

/**-------------------------------------------
 * 		Imports
 **-------------------------------------------*/
#include <assert.h>
#include "common.h"
#include "basic-types.h"

class Metriek
{
 public:
  Metriek() {};
};

/**-------------------------------------------
 * 	Clusterable Points
 **-------------------------------------------*/
class Point
{
  public:
    Point();
    virtual ~Point() {};
    virtual float  distance(Point * , Metriek * ) {return 0;};
    virtual void   simpledump(int d) {};
    virtual int    cluster_elements() { return 1; };
    virtual int    clusters_with_size(int min_size, int max_size,float &min_internal_distance, float& max_internal_distance) { assert(0); };
    virtual int    get_min_maxdepth(int &min, int &max,int depth);
    virtual float  intra_distance() {return 0;};
    virtual float  guess() {return 0;};
    void prefix(int d);
};

/**-------------------------------------------
 * 	A Position
 **-------------------------------------------*
 * A position within the similarity matrix. 
 * This is used for optimalisation purposes
 */
struct Position
{
  float distance;
  signed2 x;
  signed2 y;
  bool operator == (Position & other)
  {
    return x == other.x && y == other.y;
  }
};

/**-------------------------------------------
 * 	A Couple of points
 **-------------------------------------------*
 * This is used for optimalisation purposes
 */
class Couple: public Point
{
  int first;
  int second;
 public:
  Couple(int a, int b);
  float distance2point(int idx, Metriek* metriek);
  virtual float distance(Point * , Metriek * ) { return 0; };
  virtual void simpledump(int d);
  virtual int  clusters_with_size(int min_size, int max_size,float &min_internal_distance, float& max_internal_distance);
  virtual int  cluster_elements();
  virtual float intra_distance();
  virtual int   get_min_maxdepth(int &min, int &max,int depth);
  float distance(Couple * other, Metriek * metriek);
  float guess();
};

/**-------------------------------------------
 * 	A Cluster
 **-------------------------------------------*
 * This class should be used entirily static
 */
class Cluster
{
 public:
    // all points and clusters available
    static Point * * realcontent;
    static int       realcontentsize;
    static int       realcontenttotalsize;
    static int       type_mark;
    // the similarity matrix
    // an array of indices which are part of this cluster
    static int     * pointstocluster;
    static int       size;
    static int       totalsize;
    static inline bool isPoint(int xidx);
    
    static float * * similarity;
    static Position * * prev;
    static Position * * next;
    // the similarity matrix
    Cluster();
    static float distance(int x, int y, Metriek * metriek);
    static float distance_calculate(int x, int y, Metriek * metriek);
    static float distance_memory(int x, int y);
    // adds an index or a point to this cluster
    static void add(int idx);
    static void addPoint(Point* p);
    static void reset();
    // adds globally a new point, without inserting it in any cluster
    static int addcontent(Point * p);
    static int addpoint(Point * p);
    static int addcouple(Couple* c);
    static Couple* agglomerate(Metriek * metriek);
    void dumpConnectionMatrix();
};
#endif
