/***************************************************************************
 *  BpmDj: Free Dj Tools
 * -------------------------------------------------------------------------
 *  Copyright (C) 2004 by Werner Van Belle, see copyright.txt
 *  http://bpmdj.sourceforge.net/
 *  e-mail: werner.van.belle@itek.norut.no                                
 ***************************************************************************/

#include "version.h"

#ifdef COMPILE_ALSA

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#define ALSA_PCM_NEW_HW_PARAMS_API
#define ALSA_PCM_NEW_SW_PARAMS_API
#include "dsp-alsa.h"
#include "memory.h"
#include "inout.cpp"
#include "bpmlive.h"

#ifndef MONO8BIT
#ifndef STEREO16BIT
#error "--------------------------------------"
#error " you must specify MONBO8BIT or STEREO16BIT "
#error "--------------------------------------"
#endif
#endif

/*-------------------------------------------
 *         Dsp operations
 *-------------------------------------------*/ 
 
 /**
  * to start the reading stream we create a new thread
  */

static dsp_alsa *tostart;

static void * go(void* neglect)
{
	assert(tostart);
	tostart -> run();
	return neglect;
}

void dsp_alsa::start()
{
	pthread_t *y = Allocate(1,pthread_t);
	tostart = this;
	pthread_create(y,NULL,go,NULL);
}

void dsp_alsa::run()
{
	int err = 0;
  	err = snd_pcm_prepare(dsp);
  	if (err < 0)
    		{
      		printf("dsp: problem starting alsa device: %s\n",snd_strerror(err));
    		}
  sample_type * buf = Allocate(buffer_size, sample_type);
  while(true)
  	{
	int err = snd_pcm_readi(dsp,buf,buffer_size);
	if (err<0) 
	    {
	      if (err==-EAGAIN) continue;
	      assert(err!=-ESTRPIPE);
	      assert(err!=-EBADFD);
	      if (err==-EPIPE)
		{
		  printf("Overrun\n");
		  err = snd_pcm_prepare(dsp);
		  if (err < 0)
		    printf("cant recover from overrun: %s\n",snd_strerror(err));
		}
	    }
	else if (err > 0) buffer->push(buf,err);
	}
}

void dsp_alsa::read(sample_type *target, int length) 
{
	assert(buffer);
	buffer->pull(target,length);
}

signed8 dsp_alsa::latency()
{
/*  snd_pcm_sframes_t delay;
  int err = snd_pcm_delay(dsp,&delay);
  if (err<0)
    {
      printf("error obtaining latency:%d %s\n",err,snd_strerror(err));
      return 0;
    }
  assert(err==0);
  if (verbose)
    printf("delay = %d\n",(int)delay);
  assert(delay >= 0 && (unsigned4)delay <= buffer_size);
  return delay + filled;
  */
  return 0;
}

int dsp_alsa::get_max_read()
{
	return buffer->get_max_read();
}

dsp_alsa::dsp_alsa() : dsp_driver()
{
  dsp = NULL;
  buffer_size = 0;
  period_size = 0;
  arg_latency = "150";
}

int dsp_alsa::open(char* arg_dev, int read_time, int WAVRATE)
{
  wavrate = WAVRATE;
  int err;
  unsigned int buffer_time, period_time;
  snd_pcm_hw_params_t *hparams;
  snd_pcm_sw_params_t *sparams;

  // open device
  err = snd_pcm_open(&dsp, arg_dev, SND_PCM_STREAM_CAPTURE, SND_PCM_NONBLOCK);
  if (err < 0)
    {
      printf("dsp: opening alsa device failed : %s\n",snd_strerror(err));
      return err_dsp;
    }
  
  // set dsp parameters
  snd_pcm_sw_params_alloca(&sparams);
  snd_pcm_hw_params_alloca(&hparams);
  err = snd_pcm_hw_params_any(dsp, hparams);
  if (err<0)
    {
      printf("Broken configuration file for pcm : %s\n",snd_strerror(err));
      return err_dsp;
    }
  
  // interleaved reading and writing (in contrast to channel per channel read/write)
  err = snd_pcm_hw_params_set_access(dsp, hparams, SND_PCM_ACCESS_RW_INTERLEAVED);
  if (err<0)
    {
      printf("Unable to set interleaved access to pcm device: %s\n",snd_strerror(err));
      return err_dsp;
    }
    
  #ifdef MONO8BIT
  err = snd_pcm_hw_params_set_format(dsp, hparams, SND_PCM_FORMAT_S8);
  #endif
  
  #ifdef STEREO16BIT
  err = snd_pcm_hw_params_set_format(dsp, hparams, SND_PCM_FORMAT_S16);
  #endif
  
  if (err<0)
    {
      printf("Unable to set sample format: %s\n",snd_strerror(err));
      return err_dsp;
    }
   
  #ifdef MONO8BIT
  err = snd_pcm_hw_params_set_channels(dsp, hparams, 1);
  #endif
  
  #ifdef STEREO16BIT
  err = snd_pcm_hw_params_set_channels(dsp, hparams, 2);
  #endif
  
  if (err < 0)
    {
      printf("dsp: setting amount of channels failed : %s\n",snd_strerror(err));
      return err_dsp;
    }
  
  unsigned int q = wavrate;
  err = snd_pcm_hw_params_set_rate_near(dsp, hparams, &q, 0);
  if (err < 0)
    {
      printf("dsp: setting dsp speed (%d) failed\n",q);
      return err_dsp;
    }
    
  if (q != (unsigned4)wavrate)
    {
      printf("dsp: setting dsp speed (%d) failed, resulting rate = %d \n",WAVRATE, q);
    }

  // playing latency instellen...
  period_time = atoi(arg_latency) * 1000;
  // period_time /=2; // we do this so we can afterward use multiple periods
  buffer_time = period_time * 2;
  {
    unsigned int t = buffer_time;
    int dir = 1;
    err = snd_pcm_hw_params_set_buffer_time_near(dsp,hparams,&t,&dir);
    if (err<0)
      {
	printf("     impossible to set pcm buffer time to %i (%i): %s\n", buffer_time,t,snd_strerror(err));
	return err_dsp;
      }
    err = snd_pcm_hw_params_get_buffer_size(hparams,&buffer_size);
    if (err<0)
      {
	printf("     impossible to obtain buffer size: %s\n",snd_strerror(err));
	return err_dsp;
      }

    t = period_time;
    err = snd_pcm_hw_params_set_period_time_near(dsp, hparams,&t,&dir);
    if (err<0)
      {
	printf("     impossible to set pcm period time to %i (%i): %s\n", period_time,t,snd_strerror(err));
	return err_dsp;
      }

    err = snd_pcm_hw_params_get_period_size(hparams,&period_size,&dir);
    if (err<0)
      {
	printf("     impossible to obtain period data size: %s\n",snd_strerror(err));
	return err_dsp;
      }
    if (period_size*2 - 1>buffer_size)
      {
	printf("     impossible to allocate large enough buffers. Please decrease the latency\n");
	printf("       playbuffer size = %li\n",period_size);
	printf("       total buffer    = %li\n",buffer_size);
	return err_dsp;
      }
    dir = 0;
    /*
    err = snd_pcm_hw_params_set_periods(dsp,hparams,2,0);
    if (err < 0)
      {
	printf("    impossible to set periods to 4: %s\n",snd_strerror(err));
	return err_dsp;
      }
    */
  }

  // set the hardware parameters
  err = snd_pcm_hw_params(dsp,hparams);
  if (err < 0)
    {
      printf("unable to set hw parameters for pcm: %s\n",snd_strerror(err));
      return err_dsp;
    }
  
  // nu moeten we nog de stop & start delays goed zetten (dit zijn de software parameters)
  err = snd_pcm_sw_params_current(dsp,sparams);
  if (err < 0)
    {
      printf("unable to determine sw parameters for pcm: %s\n",snd_strerror(err));
      return err_dsp;
    }
  err = snd_pcm_sw_params_set_start_threshold(dsp,sparams,buffer_size);
  if (err < 0)
    {
      printf("unable to set start to treshold mode: %s\n",snd_strerror(err));
      return err_dsp;
    }
  err = snd_pcm_sw_params_set_avail_min(dsp,sparams,period_size);
  if (err < 0 )
    {
      printf("unable to set minimum start size: %s\n",snd_strerror(err));
      return err_dsp;
    }      
  err = snd_pcm_sw_params_set_xfer_align(dsp,sparams,1);
  if (err < 0 )
    {
      printf("unable to set transfer alignation: %s\n",snd_strerror(err));
      return err_dsp;
    }      
  err = snd_pcm_sw_params(dsp,sparams);
  if (err < 0)
    {
      printf("unable to set software parameters: %s\n",snd_strerror(err));
      return err_dsp;
    }

//  if (verbose)
//    snd_pcm_dump(dsp,input);

  // allocate buffer of correct size
  buffer = new Buffer<sample_type>(WAVRATE*read_time/1000,buffer_size);
  
  start();
  return err_none;
}

void dsp_alsa::close()
{
  snd_pcm_close(dsp);
}

#endif
