/***********************************************************
 *   Bpm Live - Live BPM counting
 *-----------------------------------------------------------
 *  Copyright (c) 2004 Werner Van Belle, see copyright.txt
 *  http://werner.onlinux.be/bpmlive/
 *  e-mail: werner.van.belle@itek.norut.no
 ***********************************************************/

#include <qapplication.h>
#include <qprogressbar.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <stdio.h>
#include <qlcdnumber.h>
#include <stdlib.h>
#include <fftw3.h>
#include <qpainter.h>
#include "dialog.h"
#include "stereo-sample2.h"
#include "bpmlive-signal.h"
#include "meter.h"
#include "dsp-alsa.h"
#include "memory.h"
#include "bpmlive.h"
#include "bpm-value.h"
#include "metronome.h"
#include "bpm-counter.h"
#include "logic.h"

Logic::Logic( QWidget* parent, const char* name, bool modal, WFlags fl )
    : BpmLive( parent, name, modal, fl )
{
	history_length = 10;
	history1 = Allocate(history_length,double);
	history_next = 0;
}

int compare_double(const void* a, const void* b)
{
	double A = *((double*)a);
	double B = *((double*)b);
	if (A<B) return -1;
	if (A>B) return 1;
	return 0;
}

void Logic::instant_tempo(double bpm1)
{
	history_next ++;
	
	history1[ history_next % history_length] = bpm1;
	
	tempo_1->display(bpm1);
	
	double tosort[history_length];
	memcpy(tosort,history1,sizeof(double)*history_length);
	qsort(tosort,history_length,sizeof(double),compare_double);
	bpm1 = round(tosort[history_length/2]*10)/10;

	balanced_1->display(bpm1);
	metronome -> set_tempo(bpm1);

	app->processEvents();
}

void Logic::open_dsp(int time)
{
	dsp1_in = new dsp_alsa();
	int err = dsp1_in -> open("hw:0", time, wavrate = 11025);
  	assert(err==err_none);
}

void Logic::init_bpmcounter()
{
	bpmcounter1 = new BpmCounter(dsp1_in);
}

void Logic::init_meter()
{
	meter1 = new Meter(dsp1_in,wavrate);
}

void Logic::init_metronome()
{
	metronome = new Metronome();
}

void Logic::done(int result)
{
	final_stop = true;
	BpmLive::done(result);
}

void Logic::start_tracing()
{
	open_dsp(11889);
	init_bpmcounter();
	init_meter();
	init_metronome();
	metronome->start();
	while(!final_stop)
		{
		double bpm1 = bpmcounter1->measure();
		int    max1 = meter1->measure();
		instant_tempo(bpm1);
   		meter_1->setProgress(max1,MAX_AMP);
		}
}
